package puppet.mobility.oracle;

import java.io.BufferedReader;
import java.io.File;
import java.io.FileNotFoundException;
import java.io.FileReader;
import java.util.ArrayList;
import java.util.Iterator;
import java.util.prefs.BackingStoreException;
import java.util.prefs.Preferences;

import org.ini4j.IniFile;


public abstract class AbstractTraceReader {

    private final static int SET_OPERATION_CODE = 0;
    private final static int GET_OPERATION_CODE = 1;    
    
    protected int nodesNumber; 
    protected String traceFilename;
    protected double maxRadius;

    /*
     * Do not access to this attribute directly. Use the getter/setter below 
     */
    private ArrayList<SimulatedEvent> events;   
    
//    protected Map<String, Set<String>> adjMatrix;
    protected BufferedReader traceFile;
    protected TraceReaderThread traceThread;

//    public AbstractTraceReader(String traceFilename){
    public AbstractTraceReader(String configurationFilename){
        
        Preferences preferenceRef;
        
        try {
            preferenceRef = new IniFile(new File(configurationFilename));
            
            this.nodesNumber = Integer.parseInt(preferenceRef.node(Constants4Mobility.REFERENCE_SECTION_IN_INIFILE).get("nodes", Constants4Mobility.DEFAULT_NODES));
            this.traceFilename = preferenceRef.node(Constants4Mobility.REFERENCE_SECTION_IN_INIFILE).get("traceFilename", Constants4Mobility.DEFAULT_TRACE_FILENAME);
            this.maxRadius = Double.parseDouble(preferenceRef.node(Constants4Mobility.REFERENCE_SECTION_IN_INIFILE).get("maxRadius", Constants4Mobility.DEFAULT_MAX_RADIUS));

        } catch (BackingStoreException e) {
            e.printStackTrace();
        }    

        try {
            this.traceFile = new BufferedReader(new FileReader (this.traceFilename));
        } catch (FileNotFoundException e) {
            e.printStackTrace();
        }

        this.traceThread =  new TraceReaderThread(this, this.traceFile);
        
        
        this.events = new ArrayList<SimulatedEvent>();
//        this.adjMatrix = Collections.synchronizedMap(new HashMap<String, Set<String>>());        
    }

    synchronized private ArrayList<SimulatedEvent> accessEventList(int operationCode, ArrayList<SimulatedEvent> events){
        switch (operationCode){
            case  SET_OPERATION_CODE :
                this.events = (ArrayList<SimulatedEvent>) events.clone();
                break;
            case GET_OPERATION_CODE :
                break;
        }
        return this.events;
    }
        
    public void setEventList(ArrayList<SimulatedEvent> events){
        this.accessEventList(SET_OPERATION_CODE, events);
    }
    
    public ArrayList<SimulatedEvent> getEventList(){
        return this.accessEventList(GET_OPERATION_CODE, null);
    } 
    
    /*
     * Implements the metric defining if 'b' is visible from 'a'.
     * This function has to implement just the logic in order to define a metric for the
     * visibility. 
     * In order to deal with visibility from external classes, do not use this function.
     * Use function the 'isVisible(String, String)' instead.
     */
    protected abstract boolean metricIsVisible (SimulatedEvent a, SimulatedEvent b);
    
    /*
     * Check if the node named 'nodeNameB' is visible from the node named 'nodeNameA'. 
     * In order to deal with visibility from external classes, use this function.
     */
    public boolean isVisible(String nodeNameA, String nodeNameB) {
      Iterator<SimulatedEvent> eventIterator = this.getEventList().iterator();
      SimulatedEvent source = null;
      SimulatedEvent destination = null;
      SimulatedEvent eventTemp;
      boolean found = false;
      while ((eventIterator.hasNext())&&(!found)){
          eventTemp = eventIterator.next();
          if (eventTemp.getNodeName().equalsIgnoreCase(nodeNameA))
              source = eventTemp;
          if (eventTemp.getNodeName().equalsIgnoreCase(nodeNameB))
              destination = eventTemp;
          found = (source != null) && (destination != null);
      }
      if (found)
          return this.metricIsVisible(source, destination);

      return false;
    } 

    /*
     * Implements the metric defining a damp percentage index from the node 
     * named 'nodeNameB' to the node named 'nodeNameA'.
     * The damp percentage index varies in [0..1]:
     *  0 -> no dump effect
     *  1 -> max dump effect 
     * In order to deal damp effect index from the external classes, do not use this function.
     * Use function the 'dampEffect(String, String)' instead.
     */
    protected abstract double metricDampEffect (SimulatedEvent a, SimulatedEvent b);

    /*
     * Returns a damp percentage index from the node 
     * named 'nodeNameB' to the node named 'nodeNameA'.
     * The damp percentage index varies in [0..1]:
     *  0 -> no dump effect
     *  1 -> max dump effect 
     */
    public double dampEffect (String nodeNameA, String nodeNameB){
        Iterator<SimulatedEvent> eventIterator = this.getEventList().iterator();
        SimulatedEvent source = null;
        SimulatedEvent destination = null;
        SimulatedEvent eventTemp;
        boolean found = false;
        
        while ((eventIterator.hasNext())&&(!found)){
            eventTemp = eventIterator.next();
            if (eventTemp.getNodeName().equalsIgnoreCase(nodeNameA))
                source = eventTemp;
            if (eventTemp.getNodeName().equalsIgnoreCase(nodeNameB))
                destination = eventTemp;
            found = (source != null) && (destination != null);
        }

        if (found)
            return this.metricDampEffect(source, destination);
        return 1;
    }

    public void startMobilitySimulation (){
        this.traceThread.start();
    }
}
